import unittest
from unittest import mock

import requests
import requests_mock

from ..client import IntrospectClient
from .requests_introspect import RequestsIntrospect


class RequestsIntrospectTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_session = mock.Mock(requests.Session)
        self.introspect_client = mock.Mock(
            wraps=IntrospectClient(self.mock_session, "http://api_url", "test"),
            api_url="http://api_url",
        )
        self.requests_introspect = RequestsIntrospect(self.introspect_client)
        self.requests_introspect.register()

    def tearDown(self) -> None:
        self.requests_introspect.unregister()

    @requests_mock.Mocker()
    def test_instrumented_request_sends_checkpoint_for_success(self, req_mock) -> None:
        req_mock.post("http://external-api", text="ok")
        got = requests.post("http://external-api", json={"key": "value"}, timeout=1)
        assert got.text == "ok"
        self.introspect_client.send.assert_called_once_with(
            {
                "location": "test/requests",
                "start_ts": mock.ANY,
                "finish_ts": mock.ANY,
                "input": '{"method": "post", "url": "http://external-api", "headers": {}, "body": null}',
                "output": '{"status_code": 200, "headers": {}, "body": "ok"}',
            }
        )

    @requests_mock.Mocker()
    def test_instrumented_request_sends_checkpoint_for_error(self, req_mock) -> None:
        req_mock.post("http://external-api/fail", status_code=500)
        got = requests.post("http://external-api/fail", json={"key": "value"}, timeout=1)
        assert got.status_code == 500
        self.introspect_client.send.assert_called_once_with(
            {
                "location": "test/requests",
                "start_ts": mock.ANY,
                "finish_ts": mock.ANY,
                "input": '{"method": "post", "url": "http://external-api/fail", "headers": {}, "body": null}',
                "output": '{"status_code": 500, "headers": {}, "body": ""}',
            }
        )

    @requests_mock.Mocker()
    def test_instrumented_request_sends_checkpoint_for_exception(self, req_mock) -> None:
        req_mock.post("http://external-api/fail", exc=requests.exceptions.ConnectTimeout)
        with self.assertRaises(requests.exceptions.ConnectTimeout):
            requests.post("http://external-api/fail", json={"key": "value"}, timeout=1)
        self.introspect_client.send.assert_called_once_with(
            {
                "location": "test/requests",
                "start_ts": mock.ANY,
                "finish_ts": mock.ANY,
                "input": '{"method": "post", "url": "http://external-api/fail", "headers": {}, "body": null}',
                "output": '{"error": ""}',
            }
        )

    @requests_mock.Mocker()
    def test_instrumented_request_skips_i8t_url(self, req_mock) -> None:
        req_mock.post("http://api_url", text="no")
        got = requests.post("http://api_url", json={"key": "value"}, timeout=1)
        assert got.text == "no"
        self.introspect_client.send.assert_not_called()



import json
import logging
import time

import requests

logger = logging.getLogger(__name__)


class IntrospectClient:
    def __init__(self, session: requests.Session, api_url: str, name: str) -> None:
        self.api_url = api_url
        self._session = session
        self._name = name

    def send(self, data: dict) -> None:
        try:
            response = self._session.post(self.api_url, json=data)
            if response.status_code != 200:
                logger.error("Failed to send checkpoint: %r", response.text)
        except Exception:  # pylint: disable=broad-except
            logger.exception("Error sending checkpoint")

    def make_checkpoint(
        self, location: str, input_data: dict, output_data: dict, start_ts: float
    ) -> dict:
        return {
            "location": f"{self._name}/{location}",
            "start_ts": start_ts,
            "finish_ts": time.time(),
            "input": json.dumps(input_data),
            "output": json.dumps(output_data),
        }



import contextlib
import fnmatch
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional
from unittest import mock

from i8t.testing.session import IntrospectSession
from i8t.testing.time_range import TimeRange

from .serde import DecoratorSerde


@dataclass
class DecoratedCase(TimeRange):
    args: List[Any]
    kwargs: Dict[str, Any]
    expected: Any
    qualname: str
    _serde = DecoratorSerde()

    @classmethod
    def from_record(cls, raw_record: dict) -> "DecoratedCase":
        record = cls._serde.deserialize(raw_record)
        return cls(
            start_ts=record["start_ts"],
            finish_ts=record["finish_ts"],
            args=record["input"]["args"],
            kwargs=record["input"]["kwargs"],
            expected=record["output"],
            qualname=record["location"].partition("/")[2],
        )

    @classmethod
    def load(
        cls, session: IntrospectSession, name: str, within: Optional[TimeRange] = None
    ) -> List["DecoratedCase"]:
        return [
            cls.from_record(record) for record in session.filter_by(DecoratorFilter(name, within))
        ]


@dataclass
class Patch:
    qualname: str
    cases: List[DecoratedCase]


@dataclass
class DecoratedPatch:
    qualname: str
    cases: List[DecoratedCase]
    mock_obj: mock.Mock
    call_args_list: List[Any]

    @classmethod
    @contextlib.contextmanager
    def activate(
        cls,
        session: IntrospectSession,
        name: str = "",
        within: Optional[TimeRange] = None,
    ) -> Iterator:
        records = session.filter_by(DecoratorFilter(name, within))
        if not records:
            yield []
        else:
            with patch_many(cls._group_by_qualname(records)) as expected_calls:
                yield expected_calls

    @staticmethod
    def _group_by_qualname(records: List[dict]) -> List[Patch]:
        by_qualname: Dict[str, List[DecoratedCase]] = {}
        for record in records:
            case = DecoratedCase.from_record(record)
            by_qualname.setdefault(case.qualname, []).append(case)
        return [Patch(qualname=qualname, cases=cases) for qualname, cases in by_qualname.items()]


@contextlib.contextmanager
def patch_many(patches: List[Patch]) -> Iterator[List[DecoratedPatch]]:
    patch = patches[0]
    side_effect = [case.expected for case in patch.cases]
    with mock.patch(patch.qualname, spec=True, side_effect=side_effect) as mocked:
        case_mock = DecoratedPatch(
            qualname=patch.qualname,
            cases=patch.cases,
            mock_obj=mocked,
            call_args_list=[mock.call(*case.args, **case.kwargs) for case in patch.cases],
        )
        if patches[1:]:
            with patch_many(patches[1:]) as expected_calls:
                yield [case_mock] + expected_calls
        else:
            yield [case_mock]


class DecoratorFilter:
    def __init__(self, name: str, within: Optional[TimeRange]) -> None:
        self._name = name
        self._within = within

    def __call__(self, record: dict) -> bool:
        return self._match_location(record["location"]) and self._match_within(record)

    def _match_location(self, location: str) -> bool:
        return fnmatch.fnmatch(location.partition("/")[2], "*" + self._name)

    def _match_within(self, record: dict) -> bool:
        if not self._within:
            return True
        return (
            record["start_ts"] >= self._within.start_ts
            and record["finish_ts"] <= self._within.finish_ts
        )



from typing import Tuple

import flask

from i8t.client import IntrospectClient


class FlaskAdapter:
    LOCATION = "flask"

    def __init__(self, client: IntrospectClient) -> None:
        self._client = client

    def record(
        self, start_time: float, request: flask.Request, response: flask.Response
    ) -> flask.Response:
        self._send(self._for_success(start_time, request, response))
        return response

    def _send(self, checkpoint_params: Tuple) -> None:
        self._client.send(self._client.make_checkpoint(*checkpoint_params))

    def _for_success(
        self, start_time: float, request: flask.Request, response: flask.Response
    ) -> Tuple:
        return (
            "flask",
            {
                "method": request.method,
                "url": request.url,
                "headers": dict(request.headers),
                "args": request.args,
                "form": request.form,
                "json": request.get_json(silent=True),
                "data": request.get_data(as_text=True),
            },
            {
                "status_code": response.status_code,
                "headers": dict(response.headers),
                "body": response.get_data(as_text=True),
            },
            start_time,
        )



import json
import unittest
from unittest import mock
from unittest.mock import patch

import flask
import requests

from ..client import IntrospectClient
from .flask_introspect import FlaskIntrospect


class TestFlaskIntrospect(unittest.TestCase):
    def setUp(self):
        self.app = flask.Flask(__name__)
        self.mock_session = mock.Mock(requests.Session)
        self.mock_client = mock.Mock(
            wraps=IntrospectClient(self.mock_session, "api_url", "test_client")
        )
        self.flask_introspect = FlaskIntrospect(self.mock_client)
        self.flask_introspect.register(self.app)

        # A simple route for testing
        @self.app.route("/test", methods=["POST"])
        def test_route():
            return "Test Response", 200

    @patch("time.time", side_effect=[1, 5, 30, 100])
    def test_before_and_after_request(self, _):
        # Use Flask's test client to simulate requests
        with self.app.test_client() as client:
            # Act: send a POST request to the /test route
            client.post("/test", data="Test Body", headers={"Test-Header": "HeaderValue"})

            # Assert: Check if the before_request correctly sets the start time
            self.assertEqual(flask.g.start_time, 1)

        # Assert: Check if the after_request triggers the introspect client
        self.mock_client.send.assert_called_once_with(
            {
                "location": "test_client/flask",
                "start_ts": 1,
                "finish_ts": 5,
                "input": json.dumps(
                    {
                        "method": "POST",
                        "url": "http://localhost/test",
                        "headers": {
                            "User-Agent": "Werkzeug/3.0.4",
                            "Host": "localhost",
                            "Content-Length": "9",
                            "Test-Header": "HeaderValue",
                        },
                        "args": {},
                        "form": {},
                        "json": None,
                        "data": "Test Body",
                    }
                ),
                "output": json.dumps(
                    {
                        "status_code": 200,
                        "headers": {
                            "Content-Type": "text/html; charset=utf-8",
                            "Content-Length": "13",
                        },
                        "body": "Test Response",
                    }
                ),
            }
        )

    @patch("time.time", side_effect=[1, 5, 30, 100])
    def test_before_request(self, _):
        with self.app.test_client() as client:
            client.get("/test")

            # Assert: start time is set in g before the request
            self.assertEqual(flask.g.start_time, 1)

    @patch("time.time", side_effect=[1, 5, 30, 100])
    def test_after_request(self, _):
        with self.app.test_client() as client:
            client.get("/test")

            # Assert: after_request sends data to the introspect client
            self.mock_client.send.assert_called_once_with(
                {
                    "location": "test_client/flask",
                    "start_ts": 1,
                    "finish_ts": 5,
                    "input": mock.ANY,
                    "output": mock.ANY,
                }
            )



import logging
import unittest
from unittest import mock

import requests

from .client import IntrospectClient, logger


class TestIntrospectClient(unittest.TestCase):
    def setUp(self) -> None:
        logger.setLevel(logging.ERROR)
        logger.addHandler(logging.StreamHandler())

    def test_send_successful(self):
        # Arrange
        mock_session = mock.Mock(requests.Session)
        mock_session.post.return_value.status_code = 200
        client = IntrospectClient(mock_session, "http://example.com", "test_client")

        # Act
        client.send({"key": "value"})

        # Assert
        mock_session.post.assert_called_once_with("http://example.com", json={"key": "value"})

    def test_send_failure(self):
        # Arrange
        mock_session = mock.Mock(requests.Session)
        mock_session.post.return_value.status_code = 500
        client = IntrospectClient(mock_session, "http://example.com", "test_client")

        # Act
        with self.assertLogs(logger=logger, level="DEBUG") as log:
            client.send({"key": "value"})

        # Assert
        self.assertIn("ERROR:i8t.client:Failed to send checkpoint", log.output[0])

    def test_send_exception(self):
        # Arrange
        mock_session = mock.Mock(requests.Session)
        mock_session.post.side_effect = Exception("Test exception")
        client = IntrospectClient(mock_session, "http://example.com", "test_client")

        # Act
        with self.assertLogs(logger=logger, level="ERROR") as log:
            client.send({"key": "value"})

        # Assert
        self.assertIn("ERROR:i8t.client:Error sending checkpoint", log.output[0])

    def test_make_checkpoint(self):
        # Arrange
        client = IntrospectClient(None, "http://example.com", "test_client")

        # Act
        with mock.patch("time.time", return_value=5):
            checkpoint = client.make_checkpoint(
                "location1", {"input": "data"}, {"output": "data"}, 1
            )

        # Assert
        self.assertEqual(
            checkpoint,
            {
                "location": "test_client/location1",
                "start_ts": 1,
                "finish_ts": 5,
                "input": '{"input": "data"}',
                "output": '{"output": "data"}',
            },
        )



import contextlib
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterator, Tuple

import requests

from i8t.client import IntrospectClient


@dataclass
class _Params:
    start_time: float
    method: str
    url: str
    kwargs: Dict[str, Any]


class RequestsAdapter:
    LOCATION = "requests"

    def __init__(self, client: IntrospectClient) -> None:
        self._client = client

    def should_record(self, url: str) -> bool:
        return url != self._client.api_url

    @contextlib.contextmanager
    def record(
        self, method: str, url: str, kwargs: Dict[str, Any]
    ) -> Iterator[Callable[[requests.Response], requests.Response]]:
        params = _Params(start_time=time.time(), method=method, url=url, kwargs=kwargs)

        def recorder(response: requests.Response) -> requests.Response:
            self._send(self._for_success(params, response))
            return response

        try:
            yield recorder
        except Exception as exc:
            self._send(self._for_exception(params, exc))
            raise

    def _send(self, checkpoint_params: Tuple) -> None:
        self._client.send(self._client.make_checkpoint(*checkpoint_params))

    def _for_success(self, params: _Params, response: requests.Response) -> Tuple:
        return (
            self.LOCATION,
            self._input(params),
            {
                "status_code": response.status_code,
                "headers": dict(response.headers),
                "body": response.text,
            },
            params.start_time,
        )

    def _for_exception(self, params: _Params, exc: Exception) -> Tuple:
        return (
            self.LOCATION,
            self._input(params),
            {"error": str(exc)},
            params.start_time,
        )

    def _input(self, params: _Params) -> dict:
        return {
            "method": params.method,
            "url": params.url,
            "headers": dict(params.kwargs.get("headers", {})),
            "body": params.kwargs.get("data", None),
        }



import unittest
from unittest import mock

import requests

from i8t.client import IntrospectClient

from .decorator_introspect import DecoratorIntrospect, introspect


class TestDecoratorIntrospect(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_session = mock.Mock(requests.Session)
        self.mock_session.post.return_value.status_code = 200
        self.decorator = DecoratorIntrospect(
            IntrospectClient(self.mock_session, "http://example.com", "test_client")
        )
        self.decorator.register()

    def tearDown(self) -> None:
        self.decorator.unregister()

    def test_function_introspected(self):
        with mock.patch("time.time", side_effect=[1000, 2000]):
            result = dummy_func(1, 2)

        self.assertEqual(result, 3)
        self.mock_session.post.assert_called_once_with(
            "http://example.com",
            json={
                "location": "test_client/i8t.instrument.test_decorator_introspect.dummy_func",
                "start_ts": 1000,
                "finish_ts": 2000,
                "input": '"fCQBt000000001elqie@VRC14luH3i0)~`~25Wa=a%XdteUx=B"',
                "output": '"fCNheE&"',
            },
        )

    def test_method_introspected(self):
        dummy = Dummy()
        with mock.patch("time.time", side_effect=[1000, 2000]):
            result = dummy.method(3)

        self.assertEqual(result, 6)
        self.mock_session.post.assert_called_once_with(
            "http://example.com",
            json={
                "location": "test_client/i8t.instrument.test_decorator_introspect.Dummy.method",
                "start_ts": 1000,
                "finish_ts": 2000,
                "input": '"fCQBr000000001elqie@VRC14luHAJl#B*zcVTj8bCi9QbuI"',
                "output": '"fCNhhE&"',
            },
        )

    def test_nested_function_introspected(self):
        dummy = Dummy()
        with mock.patch("time.time", side_effect=[1, 2, 3, 4]):
            result = dummy.nested(3)

        self.assertEqual(result, 12)
        assert self.mock_session.post.call_args_list == [
            mock.call(
                "http://example.com",
                json={
                    "location": "test_client/i8t.instrument.test_decorator_introspect.Dummy.nested.<locals>.func",
                    "start_ts": 2,
                    "finish_ts": 3,
                    "input": '"fCQBt000000001elqie@VRC14luH9k28NW325Wa=a%XdteUx=B"',
                    "output": '"fCNhkE&"',
                },
            ),
            mock.call(
                "http://example.com",
                json={
                    "location": "test_client/i8t.instrument.test_decorator_introspect.Dummy.nested",
                    "start_ts": 1,
                    "finish_ts": 4,
                    "input": '"fCQBr000000001elqie@VRC14luHAJl#B*zcVTj8bCi9QbuI"',
                    "output": '"fCNhnE&"',
                },
            ),
        ]

    def test_unregister_disables(self):
        # Disable introspect decorator and check that checkpoint is not sent
        self.decorator.unregister()
        self.assertEqual(dummy_func(1, 2), 3)
        self.mock_session.post.assert_not_called()

    def test_exception_introspected(self):
        # Act & Assert
        with mock.patch("time.time", side_effect=[1000, 2000]):
            with self.assertRaises(ValueError):
                dummy_raise(1, 2)

        self.mock_session.post.assert_called_once_with(
            "http://example.com",
            json={
                "location": "test_client/i8t.instrument.test_decorator_introspect.dummy_raise",
                "start_ts": 1000,
                "finish_ts": 2000,
                "input": '"fCQBt000000001elqie@VRC14luH3i0)~`~25Wa=a%XdteUx=B"',
                "output": '"fCQBt000000001el#B&sa&m8Sl#C8kWpi{OWq4y{aCB*JZj^H_"',
            },
        )


@introspect
def dummy_func(first, second):
    return first + second


@introspect
def dummy_raise(first, second):
    raise ValueError("Test exception")


class Dummy:
    @introspect
    def method(self, param: int) -> int:
        return param * 2

    @introspect
    def nested(self, param: int) -> int:
        @introspect
        def func(first, second):
            return first + second

        return func(param, param * 2) + param



import time

from i8t.client import IntrospectClient

from .serde import DecoratorSerde


class DecoratorExporter:
    def __init__(self, client: IntrospectClient) -> None:
        self._client = client
        self._serde = DecoratorSerde()

    def wrapper(self, func, *args, **kwargs):
        start_time = time.time()
        result = None
        try:
            result = func(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            result = {"error": str(exc)}
            raise
        finally:
            self._send(
                f"{func.__module__}.{func.__qualname__}",
                {"args": args[1:] if self._is_method(func, args) else args, "kwargs": kwargs},
                result,
                start_time,
            )
        return result

    def _is_method(self, func, args) -> bool:
        if func.__qualname__ != func.__name__:
            # Could be method or nested function or both
            # https://peps.python.org/pep-3155/
            class_name = ""
            if args:
                try:
                    class_name = args[0].__class__.__name__
                except AttributeError:
                    pass
            if func.__qualname__ == f"{class_name}.{func.__name__}":
                return True
        return False

    def _send(self, *args) -> None:
        self._client.send(self._client.make_checkpoint(*self._serde.serialize(*args)))



import base64
import pickle
from typing import Any, Tuple


class DecoratorSerde:
    def serialize(self, location, input_, output, start_time) -> Tuple:
        return (
            location,
            self._encode(input_),
            self._encode(output),
            start_time,
        )

    def deserialize(self, record: dict) -> dict:
        return dict(
            record,
            input=self._decode(record["input"]),
            output=self._decode(record["output"]),
        )

    def _decode(self, obj_str) -> Any:
        return pickle.loads(base64.b85decode(obj_str.encode("utf-8")))

    def _encode(self, obj) -> str:
        return base64.b85encode(pickle.dumps(obj)).decode("utf-8")



import fnmatch
from dataclasses import dataclass
from typing import Any, Dict, List
from urllib.parse import urlparse

import flask.testing
import werkzeug.test

from i8t.testing.session import IntrospectSession
from i8t.testing.time_range import TimeRange

from .exporter import FlaskAdapter


@dataclass
class FlaskCase(TimeRange):  # pylint: disable=too-many-instance-attributes
    url: str
    method: str
    headers: Dict[str, str]
    data: str
    form: Dict[str, Any]
    json: Dict[str, Any]
    expected_status_code: int
    expected_headers: Dict[str, str]
    expected_body: str

    @classmethod
    def from_record(cls, record: dict) -> "FlaskCase":
        return cls(
            start_ts=record["start_ts"],
            finish_ts=record["finish_ts"],
            url=record["input"]["url"],
            method=record["input"]["method"],
            headers=record["input"]["headers"],
            form=record["input"]["form"],
            json=record["input"]["json"],
            data=record["input"]["data"],
            expected_status_code=record["output"]["status_code"],
            expected_headers=record["output"]["headers"],
            expected_body=record["output"]["body"],
        )

    def call(self, client: flask.testing.FlaskClient) -> werkzeug.test.TestResponse:
        return client.open(self.url, method=self.method, json=self.json, headers=self.headers)

    @classmethod
    def load(cls, session: IntrospectSession, path: str) -> List["FlaskCase"]:
        return [cls.from_record(record) for record in session.filter_by(FlaskFilter(path))]


class FlaskFilter:
    def __init__(self, path: str) -> None:
        self._path = path

    def __call__(self, record: dict) -> bool:
        return self._match_location(record["location"]) and self._match_path(record["input"]["url"])

    def _match_location(self, location: str) -> bool:
        return location.partition("/")[2] == FlaskAdapter.LOCATION

    def _match_path(self, url: str) -> bool:
        return fnmatch.fnmatch(urlparse(url).path, self._path)



from functools import wraps
from typing import Optional

from i8t.adapters.decorator_adapter.exporter import DecoratorExporter
from i8t.client import IntrospectClient


class DecoratorIntrospect:
    instance: Optional["DecoratorIntrospect"] = None

    def __init__(self, client: IntrospectClient) -> None:
        self._client = client
        self._decorator_exporter = DecoratorExporter(client)

    def register(self) -> None:
        self.__class__.instance = self

    def unregister(self) -> None:
        self.__class__.instance = None

    def wrapper(self, func, *args, **kwargs):
        return self._decorator_exporter.wrapper(func, *args, **kwargs)


def introspect(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if DecoratorIntrospect.instance:
            return DecoratorIntrospect.instance.wrapper(func, *args, **kwargs)
        return func(*args, **kwargs)

    return wrapper



import json
import sys
import time
from typing import Iterable, Optional, Set

import requests


class CheckpointCollector:
    def __init__(self) -> None:
        self._all_records: Set[str] = set()

    def filter_new(self, latest: Iterable[str]) -> Iterable[str]:
        """Records latest lines and returns only new ones."""
        for line in latest:
            if line not in self._all_records:
                self._all_records.add(line)
                yield line


class CheckpointFetcher:
    def __init__(self, api_url: str, session: requests.Session) -> None:
        self._api_url = api_url
        self._session = session

    def fetch(self) -> Iterable[str]:
        response = self._session.get(self._api_url, timeout=1)
        for record in response.json():
            yield json.dumps(self._parse(record))

    def _parse(self, record: dict) -> dict:
        return dict(
            record,
            input=json.loads(record["input"]),
            output=json.loads(record["output"]),
        )


class CheckpointPoller:
    DEFAULT_DELAY_SEC = 1

    def __init__(
        self,
        checkpoint_fetcher: CheckpointFetcher,
        checkpoint_collector: CheckpointCollector,
        delay_sec: float = DEFAULT_DELAY_SEC,
    ) -> None:
        self._checkpoint_fetcher = checkpoint_fetcher
        self._checkpoint_collector = checkpoint_collector
        self._delay_sec = delay_sec

    def poll_once(self) -> Iterable[str]:
        yield from self._checkpoint_collector.filter_new(self._checkpoint_fetcher.fetch())

    def run_forever(self) -> Iterable[str]:
        while True:
            try:
                yield from self.poll_once()
                time.sleep(self._delay_sec)
            except KeyboardInterrupt:
                break
            except Exception as exc:  # pylint: disable=broad-except
                print(f"WARNING: {exc}", file=sys.stderr)


def main(session: Optional[requests.Session] = None) -> None:
    checkpoint_poller = CheckpointPoller(
        checkpoint_fetcher=CheckpointFetcher(
            api_url=sys.argv[1],
            session=session or requests.Session(),
        ),
        checkpoint_collector=CheckpointCollector(),
    )
    for i, line in enumerate(checkpoint_poller.run_forever()):
        print(f"\rCollected {i+1} checkpoints. Press Ctrl+C to stop.", file=sys.stderr, end="")
        print(line)


if __name__ == "__main__":
    main()



import time

import flask

from i8t.adapters.flask_adapter.exporter import FlaskAdapter
from i8t.client import IntrospectClient


class FlaskIntrospect:
    def __init__(self, client: IntrospectClient) -> None:
        self._client = client
        self._flask_adapter = FlaskAdapter(client)

    def register(self, app: flask.Flask) -> None:
        app.before_request(self.before_request)
        app.after_request(self.after_request)

    def before_request(self) -> None:
        flask.g.start_time = time.time()

    def after_request(self, response: flask.Response) -> flask.Response:
        self._flask_adapter.record(flask.g.start_time, flask.request, response)
        return response



import requests

from i8t.adapters.requests_adapter.exporter import RequestsAdapter
from i8t.client import IntrospectClient


class RequestsIntrospect:
    def __init__(self, client: IntrospectClient) -> None:
        self._client = client
        self._original_request = requests.Session.request
        self._session_request = requests.Session().request
        self._requests_adapter = RequestsAdapter(self._client)

    def register(self) -> None:
        requests.Session.request = self._instrumented_request  # type: ignore[assignment]

    def unregister(self) -> None:
        requests.Session.request = self._original_request  # type: ignore[assignment]

    def _instrumented_request(self, method, url, **kwargs):
        if not self._requests_adapter.should_record(url):
            return self._session_request(method, url, **kwargs)
        with self._requests_adapter.record(method, url, kwargs) as recorder:
            return recorder(self._session_request(method, url, **kwargs))



import json
from typing import Callable, List


class IntrospectSession:
    def __init__(self, records: List[dict]) -> None:
        self._records = records

    @classmethod
    def from_jsonl(cls, jsonl: str, main_is: str = "") -> "IntrospectSession":
        with open(jsonl, encoding="utf-8") as fobj:
            records = [json.loads(line.strip()) for line in fobj]
            for record in records:
                record["location"] = record["location"].replace("__main__", main_is)
            return cls(records)

    def filter_by(self, filter_func: Callable[[dict], bool]) -> List[dict]:
        return [record for record in self._records if filter_func(record)]



import contextlib
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional

import requests_mock

from i8t.testing.session import IntrospectSession
from i8t.testing.time_range import TimeRange

from .exporter import RequestsAdapter


@dataclass
class RequestsMock:
    method: str
    url: str
    text: str
    headers: Dict[str, str]
    status_code: int

    @classmethod
    def from_record(cls, record: dict) -> "RequestsMock":
        return cls(
            method=record["input"]["method"],
            url=record["input"]["url"],
            text=record["output"]["body"],
            headers=record["output"]["headers"],
            status_code=record["output"]["status_code"],
        )

    def register(self, mocker: requests_mock.Mocker) -> None:
        mocker.register_uri(
            self.method,
            self.url,
            text=self.text,
            headers=self.headers,
            status_code=self.status_code,
        )

    @classmethod
    @contextlib.contextmanager
    def activate(cls, session: IntrospectSession, within: Optional[TimeRange] = None) -> Iterator:
        with requests_mock.Mocker() as mocker:
            for case in cls.load(session, within):
                case.register(mocker)
            yield

    @classmethod
    def load(cls, session: IntrospectSession, within: Optional[TimeRange]) -> List["RequestsMock"]:
        return [cls.from_record(record) for record in session.filter_by(RequestsFilter(within))]


class RequestsFilter:
    def __init__(self, within: Optional[TimeRange]) -> None:
        self._within = within

    def __call__(self, record: dict) -> bool:
        return self._match_location(record["location"]) and self._match_within(record)

    def _match_location(self, location: str) -> bool:
        return location.partition("/")[2] == RequestsAdapter.LOCATION

    def _match_within(self, record: dict) -> bool:
        if not self._within:
            return True
        return (
            record["start_ts"] >= self._within.start_ts
            and record["finish_ts"] <= self._within.finish_ts
        )
